clc
close all
clear all

global mdl
global xdata
global ydata
global idxX
global pwrs

warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data1.xlsx');
delete("data1b_*.txt");
outFile = strcat("data1b_",dtstr,".txt");
diary(outFile)

fprintf("Program  clem1b\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');
[maxrows,~] = size(xyData);
ydata = xyData(:,1);
xdata = xyData(:,2:end);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);
[n,nXVars] = size(xdata);
numIndpX = xyTransf(3,1);
idxX = zeros(1,numIndpX);
pwrs = zeros(1,numIndpX);
for i=1:numIndpX
  idxX(i) = xyTransf(3,1+i);
  pwrs(i) = xyTransf(4,1+i);
  if isnan(pwrs(i)), pwrs(i) = 1; end
end

Lb = xyTransf(1,:);
Ub = xyTransf(2,:);
nVars = length(Ub);
iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx= Lb;
bStop = false;
iter = 0;
fprintf("------------- Calculating Max Iters ----------------\n");
while ~bStop 
  iter = iter +1;
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
str = num2str(iter);
str = fliplr(regexprep(fliplr(str), '(\d+\.)?(\d{3})(?=\S+)', '$1$2,'));
fprintf("Max iters = %s\n", str);
MaxIters = iter;
FivePercent = fix(iter/20);

iFrom = Lb;
iTo = Ub;
iStep = 1 + zeros(1,nVars);
idx= Lb;
bestX = zeros(1,nVars);
bestFx = 0;
% --------------------------------- start main loop
bStop = false;
iter = 0;
fprintf("------------- Maximizing Fx ----------------\n");
while ~bStop 
  iter = iter +1;
  if mod(iter, FivePercent) == 0
    percent = 100*iter/MaxIters;
    fprintf("Progress %5.2f%%\n", percent);
  end

  r2adj = myFit(idx);
  if r2adj > bestFx
    bestFx = r2adj;
    bestX = idx;
    bestMdl = mdl;
    fprintf("Iter %d, Fx = %e, X=[", iter, bestFx)
    fprintf("%d, ", bestX);
    fprintf("]\n");
  end
  % Start implementing the quasi–nested loops
  idx(1) = idx(1) + iStep(1);
  if idx(1) > iTo(1)
    idx(1) = iFrom(1);
    for i = 2:nVars
      idx(i) = idx(i) + iStep(i);
      if idx(i) > iTo(i) && i < nVars
        idx(i) = iFrom(i);
      elseif idx(i) > iTo(i) && i == nVars
        bStop = true;
      else
        break
      end
    end
  end
end
fprintf("Regression model\n")
format long
AdjR2 = bestMdl.Rsquared.Adjusted;
fprintf("Adjusted R-square = %14.12f\n", AdjR2);
mdl = bestMdl
bx = bestX;
bestX = [AdjR2];
for i = 1:numIndpX
    bestX = [bestX idxX(i)];
end
bestX = [bestX bx];
k =length(bestX);
res = zeros(3,k);
res(1,:) = bestX;
k2 = 1+numIndpX;
res(2,2:k2+2) = mdl.Coefficients{:,1}';
res(3,3:2+numIndpX) = pwrs;
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:k2+1), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
writematrix(res(3,3:2+numIndpX), filename, 'Sheet', 'Results', 'Range', 'B4:Z4');
% 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));

i =2+numIndpX;
j=i+numIndpX;
sJ = num2str(bestX(j));
if bestX(i) == 0
  sx1 = strcat("ln(X", sJ,")");
elseif bestX(i) == 1
  sx1 = strcat("X", sJ);
elseif bestX(i) == 4
  sx1 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx1 = strcat("X", sJ,"^", num2str(bestX(i)));
end
ix1 = sJ;
k2=1+numIndpX+2;
if bestX(k2+1) == 1
  sOp1 = "*";
else
  sOp1 = "/";
end
if bestX(k2+2) == 1
  sOp2 = "*";
else
  sOp2 = "/";
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx2 = strcat("X", sJ);
elseif bestX(i) == 0
  sx2 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx2 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx2 = strcat("X", sJ,"^", num2str(bestX(i)));
end
ix2 = sJ;
s = strcat("Y = (", num2str(mdl.Coefficients{1,1}),")");
fprintf("\n\nBest model is:\n%s\n", s)
for i=1:numIndpX
  j = idxX(i);
  sJ= num2str(j);
  if pwrs(i)==1
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ);
  elseif pwrs(i)==0
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*ln(X", sJ,")");
  else
    sp = num2str(pwrs(i));
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ,"^",sp);
  end
  fprintf("%s\n", s)
end
s = strcat("   + (",num2str(mdl.Coefficients{2+numIndpX,1}), ")*", sx1, sOp1, sx2);
fprintf("%s\n", s)


format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==2
    x = x.^2;
  elseif p==3
    x = x.^3;
  elseif p==4
    x = sqrt(x);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata
  global idxX
  global pwrs

  numIndpX = length(idxX);
  y = ydata;
  px1 = coeff(1);
  px2 = coeff(2);
  i1 = coeff(3);
  i2 = coeff(4);
  op1 = coeff(5);

  if i1==i2
    r = 0;
    return;
  end
   
  x1 = myfx(xdata(:,i1),px1);
  x2 = myfx(xdata(:,i2),px2);
  if op1==1
    x12 = x1.*x2;
  else
    x12 = x1./x2;
  end
  X = [];
  for i=1:numIndpX
    X = [X myfx(xdata(:,idxX(i)),pwrs(i))];
  end
  X = [X x12];
  mdl = fitlm(X,y);
  r = mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata
  global idxX
  global pwrs

  numIndpX = length(idxX);
  shift = numIndpX;
  px1 = res(1,2+shift);
  px2 = res(1,3+shift);
  i1 = res(1,4+shift);
  i2 = res(1,5+shift);
  op1 = res(1,6+shift);

  if px1 == 0
    x1= log(xdata(:,i1));
  elseif px1 == 1
    x1 = xdata(:,i1);
  elseif px1 == 2
    x1 = xdata(:,i1).^2;
  elseif px1 == 3
    x1 = xdata(:,i1).^3;
  elseif px1 == 4
    x1 = sqrt(xdata(:,i1));
  end      

  if px2 == 0
    x2= log(xdata(:,i2));
  elseif px2 == 1
    x2 = xdata(:,i2);
  elseif px2 == 2
    x2 = xdata(:,i2).^2;
  elseif px2 == 3
    x2 = xdata(:,i2).^3;
  elseif px2 == 4
    x2 = sqrt(xdata(:,i2));
  end   

  ycalc = res(2,2);
  for i=1:numIndpX
    ycalc = ycalc + res(2,2+i) * myfx(xdata(:,idxX(i)),pwrs(i));
  end
  k = 1 + numIndpX;
  if op1==1
    ycalc = ycalc + res(2,k+2) * x1.*x2;
  else
    ycalc = ycalc + res(2,k+2) * x1./x2;
  end
end