clc
close all
clear all

global mdl
global xdata
global ydata
global idxX
global pwrs
  
warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data3.xlsx');
delete("data3_*.txt");
outFile = strcat("data3_",dtstr,".txt");
diary(outFile)

fprintf("Program  clem3\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');
[maxrows,~] = size(xyData);
ydata = xyData(:,1);
xdata = xyData(:,2:end);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);
[n,nXVars] = size(xdata);
numIndpX = xyTransf(3,1);
idxX = zeros(1,numIndpX);
pwrs = zeros(1,numIndpX);
for i=1:numIndpX
  idxX(i) = xyTransf(3,1+i);
  pwrs(i) = xyTransf(4,1+i);
  if isnan(pwrs(i)), pwrs(i) = 1; end
end

Lb = xyTransf(1,:);
Ub = 1+xyTransf(2,:);
nVars = length(Ub);
MaxIters = 4000;
MaxPop= 400;
itersToReset = 400;
MaxTries = 5;
numTry = 0;
while true
  numTry = numTry + 1;
  if numTry > MaxTries
    fprintf("Optimization failed\n");
    return;
  end
  if numTry > 1
    fprintf("Optimization attempt # %d\n", numTry)
  end
  [bestX,gBestCost,bestMdl] = pso(@myFit,Lb,Ub,MaxPop,MaxIters,itersToReset);
  px1 = bestX(1);
  px2 = bestX(2);
  px3 = bestX(3);
  px4 = bestX(4);
  px5 = bestX(5);
  px6 = bestX(6);
  
  i1= bestX(1+numIndpX);
  i2= bestX(2+numIndpX);
  i3= bestX(3+numIndpX);
  i4= bestX(4+numIndpX);
  i5= bestX(5+numIndpX);
  i6= bestX(6+numIndpX);
  
  if (i1 == i2) || (i3 == i4) || (i5 == i6), continue, end
  % compare X1,X2 with X3,X4
  if ((i1==i3) && (px1==px3)) || ((i2==i4) && (px2==px4)), continue, end  
  if ((i1==i4) && (px1==px4)) || ((i2==i3) && (px2==px3)), continue, end
  % compare X1,X2 with X5,X6
  if ((i1==i5) && (px1==px5)) || ((i2==i6) && (px2==px6)), continue, end  
  if ((i1==i6) && (px1==px6)) || ((i2==i5) && (px2==px5)), continue, end   
  % compare X3,X4 with X5,X6
  if ((i3==i5) && (px3==px5)) || ((i4==i6) && (px4==px6)), continue, end    
  if ((i3==i6) && (px3==px6)) || ((i4==i5) && (px4==px5)), continue, end
  break;
end
mdl = bestMdl;
AdjR2 = mdl.Rsquared.Adjusted;
fprintf("Regression model\n")
fprintf("Adjusted R-square = %14.10f\n", AdjR2);
format long
mdl
bx = bestX;
bestX = [AdjR2];
for i = 1:numIndpX
    bestX = [bestX idxX(i)];
end
bestX = [bestX bx];
k =length(bestX);
res = zeros(3,k);
res(1,:) = bestX;
k2 = 4+numIndpX;
res(2,2:k2+1) = mdl.Coefficients{:,1}';
res(3,3:2+numIndpX) = pwrs;
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:k2+1), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
writematrix(res(3,3:2+numIndpX), filename, 'Sheet', 'Results', 'Range', 'B4:Z4');
% 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));

k2=length(bestX);
if bestX(k2-2) == 1
  sOp1 = "*";
else
  sOp1 = "/";
end
if bestX(k2-1) == 1
  sOp2 = "*";
else
  sOp2 = "/";
end
if bestX(k2) == 1
  sOp3 = "*";
else
  sOp3 = "/";
end

i=2+numIndpX;
j=i+numIndpX;
sJ = num2str(bestX(j));
if bestX(i) == 0
  sx1 = strcat("ln(X", sJ,")");
elseif bestX(i) == 1
  sx1 = strcat("X", sJ);
elseif bestX(i) == 4
  sx1 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx1 = strcat("X", sJ,"^", num2str(bestX(i)));
end


i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx2 = strcat("X", sJ);
elseif bestX(i) == 0
  sx2 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx2 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx2 = strcat("X", sJ,"^", num2str(bestX(i)));
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx3 = strcat("X", sJ);
elseif bestX(i) == 0
  sx3 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx3 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx3 = strcat("X", sJ,"^", num2str(bestX(i)));
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx4 = strcat("X", sJ);
elseif bestX(i) == 0
  sx4 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx4 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx4 = strcat("X", sJ,"^", num2str(bestX(i)));
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx5 = strcat("X", sJ);
elseif bestX(i) == 0
  sx5 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx5 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx5 = strcat("X", sJ,"^", num2str(bestX(i)));
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx6 = strcat("X", sJ);
elseif bestX(i) == 0
  sx6 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx6 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx6 = strcat("X", sJ,"^", num2str(bestX(i)));
end

s = strcat("Y = (", num2str(mdl.Coefficients{1,1}),")");
fprintf("\n\nBest model is:\n%s\n", s)
for i=1:numIndpX
  j = idxX(i);
  sJ= num2str(j);
  if pwrs(i)==1
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ);
  elseif pwrs(i)==0
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*ln(X", sJ,")");
  else
    sp = num2str(pwrs(i));
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ,"^",sp);
  end
  fprintf("%s\n", s)
end
s = strcat("   + (",num2str(mdl.Coefficients{2+numIndpX,1}), ")*", sx1, sOp1, sx2);
fprintf("%s\n", s)
s = strcat("   + (",num2str(mdl.Coefficients{3+numIndpX,1}), ")*", sx3, sOp2, sx4);
fprintf("%s\n", s)
s = strcat("   + (", num2str(mdl.Coefficients{4+numIndpX,1}), ")*", sx5, sOp3, sx6);
fprintf("%s\n", s)

format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==2
    x = x.^2;
  elseif p==3
    x = x.^3;
  elseif p==4
    x = sqrt(x);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata
  global idxX
  global pwrs

  [~,nXVars] = size(xdata);
  numIndpX = length(idxX);
  y = ydata;
  px1 = coeff(1);
  px2 = coeff(2);
  px3 = coeff(3);
  px4 = coeff(4);  
  px5 = coeff(5);
  px6 = coeff(6);  
  i1 = coeff(7);
  i2 = coeff(8);
  i3 = coeff(9);
  i4 = coeff(10);
  i5 = coeff(11);
  i6 = coeff(12);
  op1 = coeff(13);
  op2 = coeff(14);
  op3 = coeff(15);
  if i1 > nXVars || i2 > nXVars || i3 > nXVars || ...
      i4 > nXVars || i5 > nXVars || i6 > nXVars
    r = 1;
    return
  end

  if (i1==i2) || (i3==i4) || (i5==i6)
    r = 1;
    return;
  end
  % compare X1,X2 with X3,X4
  if ((i1==i3) && (px1==px3)) || ((i2==i4) && (px2==px4))
    r = 1;
    return;
  end     
  if ((i1==i4) && (px1==px4)) || ((i2==i3) && (px2==px3))
    r = 1;
    return;
  end     
  % compare X1,X2 with X5,X6
  if ((i1==i5) && (px1==px5)) || ((i2==i6) && (px2==px6))
    r = 1;
    return;
  end     
  if ((i1==i6) && (px1==px6)) || ((i2==i5) && (px2==px5))
    r = 1;
    return;
  end     
  % compare X3,X4 with X5,X6
  if ((i3==i5) && (px3==px5)) || ((i4==i6) && (px4==px6))
    r = 1;
    return;
  end     
  if ((i3==i6) && (px3==px6)) || ((i4==i5) && (px4==px5))
    r = 1;
    return;
  end     
   
  x1 = myfx(xdata(:,i1),px1);
  x2 = myfx(xdata(:,i2),px2);
  x3 = myfx(xdata(:,i3),px3);
  x4 = myfx(xdata(:,i4),px4);
  x5 = myfx(xdata(:,i5),px5);
  x6 = myfx(xdata(:,i6),px6);  
  if op1==1
    x12 = x1.*x2;
  else
    x12 = x1./x2;
  end
  if op2==1
    x34 = x3.*x4;
  else
    x34  = x3./x4;
  end
  if op3==1
    x56 = x5.*x6;
  else
    x56  = x5./x6;
  end  
  X = [];
  for i=1:numIndpX
    X = [X myfx(xdata(:,idxX(i)),pwrs(i))];
  end
  X = [X x12 x34 x56];
  mdl = fitlm(X,y);
  r = 1 - mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata
  global idxX
  global pwrs

  numIndpX = length(idxX);
  shift = 1 + numIndpX;
  px1 = res(1,1+shift);
  px2 = res(1,2+shift);
  px3 = res(1,3+shift);
  px4 = res(1,4+shift);
  px5 = res(1,5+shift);
  px6 = res(1,6+shift);  
  i1 = res(1,7+shift);
  i2 = res(1,8+shift);
  i3 = res(1,9+shift);
  i4 = res(1,10+shift); 
  i5 = res(1,11+shift);
  i6 = res(1,12+shift);
  op1 = res(1,13+shift);
  op2 = res(1,14+shift);
  op3 = res(1,15+shift); 
  if px1 == 0
    x1= log(xdata(:,i1));
  elseif px1 == 1
    x1 = xdata(:,i1);
  elseif px1 == 2
    x1 = xdata(:,i1).^2;
  elseif px1 == 3
    x1 = xdata(:,i1).^3;
  elseif px1 == 4
    x1 = sqrt(xdata(:,i1));
  end      

  if px2 == 0
    x2= log(xdata(:,i2));
  elseif px2 == 1
    x2 = xdata(:,i2);
  elseif px2 == 2
    x2 = xdata(:,i2).^2;
  elseif px2 == 3
    x2 = xdata(:,i2).^3;
  elseif px2 == 4
    x2 = sqrt(xdata(:,i2));
  end   

  if px3 == 0
    x3= log(xdata(:,i3));
  elseif px3 == 1
    x3 = xdata(:,i3);
  elseif px3 == 2
    x3 = xdata(:,i3).^2;
  elseif px3 == 3
    x3 = xdata(:,i3).^3;
  elseif px3 == 4
    x3 = sqrt(xdata(:,i3));
  end  

  if px4 == 0
    x4= log(xdata(:,i4));
  elseif px4 == 1
    x4 = xdata(:,i4);
  elseif px4 == 2
    x4 = xdata(:,i4).^2;
  elseif px4 == 3
    x4 = xdata(:,i4).^3;
  elseif px4 == 4
    x4 = sqrt(xdata(:,i4));
  end  

  if px5 == 0
    x5= log(xdata(:,i5));
  elseif px5 == 1
    x5 = xdata(:,i5);
  elseif px5 == 2
    x5 = xdata(:,i5).^2;
  elseif px5 == 3
    x5 = xdata(:,i5).^3;
  elseif px5 == 4
    x5 = sqrt(xdata(:,i5));
  end  

  if px6 == 0
    x6= log(xdata(:,i6));
  elseif px6 == 1
    x6 = xdata(:,i6);
  elseif px6 == 2
    x6 = xdata(:,i6).^2;
  elseif px6 == 3
    x6 = xdata(:,i6).^3;
  elseif px6 == 4
    x6 = sqrt(xdata(:,i6));
  end  

  ycalc = res(2,2);
  for i=1:numIndpX
    ycalc = ycalc + res(2,2+i) * myfx(xdata(:,idxX(i)),pwrs(i));
  end
  k = 1 + numIndpX;
  if op1==1 
    ycalc = ycalc + res(2,k+2) * x1.*x2;
  else
    ycalc = ycalc + res(2,k+2) * x1./x2;
  end
  if op2==1
    ycalc = ycalc + res(2,k+3) * x3.*x4;
  else
    ycalc = ycalc + res(2,k+3) * x3./x4;
  end
  if op3==1
    ycalc = ycalc + res(2,k+4) * x5.*x6;
  else
    ycalc = ycalc + res(2,k+4) * x5./x6;
  end
end