clc
close all
clear all

global mdl
global xdata
global ydata
global idxX
global pwrs
  
warning("off")
dt = datetime('now','TimeZone','local','Format','y-MMM-d HH_mm_ss');
dtstr = string(dt);
filename = strcat(pwd,'\data4.xlsx');
delete("data4_*.txt");
outFile = strcat("data4_",dtstr,".txt");
diary(outFile)

fprintf("Program  clem4\n\n");
fprintf("%s\n", dtstr);
fprintf("Please wait ...\n\n");
fprintf("Excel file used is %s\n", filename);
fprintf("Diary file used is %s\n", outFile);
xyData = readmatrix(filename,'Sheet','Data');
[maxrows,~] = size(xyData);
ydata = xyData(:,1);
xdata = xyData(:,2:end);
xyTransf = readmatrix(filename,'Sheet','Transf');
xyTransf = xyTransf(:,2:end);
[n,nXVars] = size(xdata);
numIndpX = xyTransf(3,1);
idxX = zeros(1,numIndpX);
pwrs = zeros(1,numIndpX);
for i=1:numIndpX
  idxX(i) = xyTransf(3,1+i);
  pwrs(i) = xyTransf(4,1+i);
  if isnan(pwrs(i)), pwrs(i) = 1; end
end

Lb = xyTransf(1,:);
Ub = 1+xyTransf(2,:);
nVars = length(Ub);
MaxIters = 3000;
MaxPop= 300;
itersToReset = 300;
MaxTries = 5;
numTry = 0;
while true
  numTry = numTry + 1;
  if numTry > MaxTries
    fprintf("Optimization failed\n");
    return;
  end
  if numTry > 1
    fprintf("Optimization attempt # %d\n", numTry)
  end
  [bestX,gBestCost,bestMdl] = pso(@myFit,Lb,Ub,MaxPop,MaxIters,itersToReset);
  px1 = bestX(1);
  px2 = bestX(2);
  px3 = bestX(3);

  i1= bestX(1+numIndpX);
  i2= bestX(2+numIndpX);
  i3= bestX(3+numIndpX);

  
  if (i1 == i2) || (i2 == i3), continue, end
  % compare X1,X2 with X3,X4
  if ((i1==i3) && (px1==px3)) || ((i2==i3) && (px2==px3)), continue, end  
  break;
end
mdl = bestMdl;
AdjR2 = mdl.Rsquared.Adjusted;
fprintf("Regression model\n")
fprintf("Adjusted R-square = %14.10f\n", AdjR2);
format long
mdl
bx = bestX;
bestX = [AdjR2];
for i = 1:numIndpX
    bestX = [bestX idxX(i)];
end
bestX = [bestX bx];
k =length(bestX);
res = zeros(3,k);
res(1,:) = bestX;
k2 = 2+numIndpX;
res(2,2:k2+1) = mdl.Coefficients{:,1}';
res(3,3:2+numIndpX) = pwrs;
writematrix(res(1,:), filename, 'Sheet', 'Results', 'Range', 'A2:Z2');
writematrix(res(2,2:k2+1), filename, 'Sheet', 'Results', 'Range', 'B3:Z3');
writematrix(res(3,3:2+numIndpX), filename, 'Sheet', 'Results', 'Range', 'B4:Z4');
% 
% now do the % projections
ycalc = project(res);
percErr = 100.*(ycalc-ydata)./ydata;
ProjMat = [ydata xdata ycalc percErr];
writematrix(ProjMat, filename, 'Sheet', 'Project', 'Range', strcat("A2:Z", num2str(1+length(ydata))));

k2=length(bestX);
if bestX(k2-1) == 1
  sOp1 = "*";
else
  sOp1 = "/";
end
if bestX(k2) == 1
  sOp2 = "*";
else
  sOp2 = "/";
end

i=2+numIndpX;
j=i+numIndpX;
sJ = num2str(bestX(j));
if bestX(i) == 0
  sx1 = strcat("ln(X", sJ,")");
elseif bestX(i) == 1
  sx1 = strcat("X", sJ);
elseif bestX(i) == 4
  sx1 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx1 = strcat("X", sJ,"^", num2str(bestX(i)));
end


i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx2 = strcat("X", sJ);
elseif bestX(i) == 0
  sx2 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx2 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx2 = strcat("X", sJ,"^", num2str(bestX(i)));
end

i=i+1;
j=j+1;
sJ = num2str(bestX(j));
if bestX(i) == 1
  sx3 = strcat("X", sJ);
elseif bestX(i) == 0
  sx3 = strcat("ln(X", sJ,")");
elseif bestX(i) == 4
  sx3 = strcat("sqrt(X", sJ,")");
elseif bestX(i) > 1
  sx3 = strcat("X", sJ,"^", num2str(bestX(i)));
end


s = strcat("Y = (", num2str(mdl.Coefficients{1,1}),")");
fprintf("\n\nBest model is:\n%s\n", s)
for i=1:numIndpX
  j = idxX(i);
  sJ= num2str(j);
  if pwrs(i)==1
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ);
  elseif pwrs(i)==0
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*ln(X", sJ,")");
  else
    sp = num2str(pwrs(i));
    s = strcat("   + (", num2str(mdl.Coefficients{1+i,1}),")*X", sJ,"^",sp);
  end
  fprintf("%s\n", s)
end
s = strcat("   + (",num2str(mdl.Coefficients{2+numIndpX,1}), ")*", sx1, sOp1, sx2, sOp2, sx3);
fprintf("%s\n", s)

format short
fprintf("Done!\n");
diary off
beep on
for i=1:3
  beep;
  pause(1);
end

function x = myfx(x,p)
  if p==0
    x = log(x);
  elseif p==2
    x = x.^2;
  elseif p==3
    x = x.^3;
  elseif p==4
    x = sqrt(x);
  end
end

function r = myFit(coeff)
  global mdl
  global xdata
  global ydata
  global idxX
  global pwrs

  [~,nXVars] = size(xdata);
  numIndpX = length(idxX);
  y = ydata;
  px1 = coeff(1);
  px2 = coeff(2);
  px3 = coeff(3);
  i1 = coeff(4);
  i2 = coeff(5);
  i3 = coeff(6);
  op1 = coeff(7);
  op2 = coeff(8);
  if i1 > nXVars || i2 > nXVars || i3 > nXVars 
    r = 1;
    return
  end

  if (i1==i2) || (i1==i3) || (i2==i3)
    r = 1;
    return;
  end
  % compare X1,X2 with X3,X4
  if ((i1==i2) && (px1==px2)) || ((i1==i3) && (px1==px3)) || ((i2==i3) && (px2==px3))
    r = 1;
    return;
  end     
   
  x1 = myfx(xdata(:,i1),px1);
  x2 = myfx(xdata(:,i2),px2);
  x3 = myfx(xdata(:,i3),px3);
  if op1==1
    x12 = x1.*x2;
  else
    x12 = x1./x2;
  end
  if op2==1
    x123 = x12.*x3;
  else
    x123  = x12./x3;
  end
  X = [];
  for i=1:numIndpX
    X = [X myfx(xdata(:,idxX(i)),pwrs(i))];
  end
  X = [X x123 ];
  mdl = fitlm(X,y);
  r = 1 - mdl.Rsquared.Adjusted;
end

function ycalc = project(res)
  global xdata
  global idxX
  global pwrs

  numIndpX = length(idxX);
  shift = 1 + numIndpX;
  px1 = res(1,1+shift);
  px2 = res(1,2+shift);
  px3 = res(1,3+shift);
  i1 = res(1,4+shift);
  i2 = res(1,5+shift);
  i3 = res(1,5+shift);
  op1 = res(1,7+shift);
  op2 = res(1,8+shift);
 
  if px1 == 0
    x1= log(xdata(:,i1));
  elseif px1 == 1
    x1 = xdata(:,i1);
  elseif px1 == 2
    x1 = xdata(:,i1).^2;
  elseif px1 == 3
    x1 = xdata(:,i1).^3;
  elseif px1 == 4
    x1 = sqrt(xdata(:,i1));
  end      

  if px2 == 0
    x2= log(xdata(:,i2));
  elseif px2 == 1
    x2 = xdata(:,i2);
  elseif px2 == 2
    x2 = xdata(:,i2).^2;
  elseif px2 == 3
    x2 = xdata(:,i2).^3;
  elseif px2 == 4
    x2 = sqrt(xdata(:,i2));
  end   

  if px3 == 0
    x3= log(xdata(:,i3));
  elseif px3 == 1
    x3 = xdata(:,i3);
  elseif px3 == 2
    x3 = xdata(:,i3).^2;
  elseif px3 == 3
    x3 = xdata(:,i3).^3;
  elseif px3 == 4
    x3 = sqrt(xdata(:,i3));
  end  

  if op1==1
    x12 = x1.*x2;
  else
    x12 = x1./x2;
  end
  if op2==1
    x123 = x12.*x3;
  else
    x123  = x12./x3;
  end


  ycalc = res(2,2);
  for i=1:numIndpX
    ycalc = ycalc + res(2,2+i) * myfx(xdata(:,idxX(i)),pwrs(i));
  end
  k = 1 + numIndpX; 
  ycalc = ycalc + res(2,end) * x123;
end